package de.uni_passau.fim.seibt.gitwrapper.process;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.io.IOUtils;

import static de.uni_passau.fim.seibt.gitwrapper.repo.GitWrapper.EXIT_FAIL;
import static java.util.logging.Level.WARNING;

/**
 * Contains static methods for executing commands and gathering their output and exit code.
 */
public class ProcessExecutor {

    private static final Logger LOG = Logger.getLogger(ProcessExecutor.class.getCanonicalName());

    /**
     * The result of a process execution.
     */
    public static class ExecRes {
        public int exitCode;
        public String stdOut;
        public String stdErr;
    }

    /**
     * Executes the given <code>command</code>. If <code>command</code> is <code>null</code> an empty {@link Optional}
     * will be returned.
     *
     * @param workingDir
     *         the working directory for the execution
     * @param redirectError
     *         whether to merge the standard output and error streams
     * @param command
     *         the command to be executed
     * @return the result of the execution or an empty {@link Optional} if there is an exception executing the command
     */
    public static Optional<ExecRes> exec(File workingDir, boolean redirectError, List<String> command) {

        if (command == null) {
            return Optional.empty();
        }

        ProcessBuilder builder = new ProcessBuilder();

        builder.command(command);
        builder.directory(workingDir);
        builder.redirectErrorStream(redirectError);

        return exec(builder);
    }

    /**
     * Executes the command represented by the given <code>builder</code>.
     *
     * @param builder
     *         the {@link ProcessBuilder} to be executed
     * @return the result of the execution or an empty {@link Optional} if there is an exception executing the command
     */
    public static Optional<ExecRes> exec(ProcessBuilder builder) {
        String cmd = String.join(" ", builder.command());
        LOG.fine(() -> String.format("Executing '%s' in '%s'.", cmd, builder.directory().getAbsolutePath()));

        try {
            ExecRes res = new ExecRes();
            Process p = builder.start();

            res.stdOut = IOUtils.toString(p.getInputStream(), StandardCharsets.UTF_8).trim();
            res.stdErr = IOUtils.toString(p.getErrorStream(), StandardCharsets.UTF_8).trim();

            IOUtils.closeQuietly(p.getInputStream());
            IOUtils.closeQuietly(p.getErrorStream());
            IOUtils.closeQuietly(p.getOutputStream());

            try {
                res.exitCode = p.waitFor();
            } catch (InterruptedException e) {
                LOG.log(WARNING, e, () -> String.format("Interrupted while waiting for '%s' to finish.", cmd));

                p.destroyForcibly();
                res.exitCode = EXIT_FAIL;
            }

            LOG.fine(() -> String.format("Execution of '%s' returned exit code %d.", cmd, res.exitCode));

            if (res.stdOut.isEmpty()) {
                LOG.finest(() -> String.format("Execution of '%s' returned no standard output.", cmd));
            } else {
                LOG.finest(() -> String.format("Execution of '%s' returned standard output:%n%s", cmd, res.stdOut));
            }

            if (res.stdErr.isEmpty()) {
                LOG.finest(() -> String.format("Execution of '%s' returned no standard error output.", cmd));
            } else {
                LOG.finest(() -> String.format("Execution of '%s' returned standard error output:%n%s", cmd, res.stdErr));
            }

            return Optional.of(res);
        } catch (IOException e) {
            LOG.log(Level.SEVERE, e, () -> String.format("Exception executing '%s'.", cmd));
            return Optional.empty();
        }
    }

    /**
     * Returns a modifiable {@link LinkedList} containing the given elements.
     *
     * @param ts
     *         the elements to return in a {@link LinkedList}
     * @param <T>
     *         the type of the elements
     * @return a {@link LinkedList} containing the elements
     */
    @SafeVarargs
    public static <T> LinkedList<T> asList(T... ts) {
        LinkedList<T> l = new LinkedList<>(); // we expect things to be inserted in front and back

        if (ts == null) {
            return l;
        }

        Collections.addAll(l, ts);

        return l;
    }
}
